# -*- coding: utf-8 -*-

import numpy as np

from ...Model import Model


class BernoulliNB(Model):
    """
    See also
    --------
    ...
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'java': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'arr[]':    '{type}[] {name} = {{{values}}};',
            'arr[][]':  '{type}[][] {name} = {{{values}}};',
            'indent':   '    ',
        },
    }
    # @formatter:on

    def __init__(self, language='java', method_name='predict',
                 class_name='Tmp'):
        super(BernoulliNB, self).__init__(language, method_name, class_name)

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : GaussianNB
            An instance of a trained GaussianNB classifier.
        """
        super(BernoulliNB, self).port(model)

        # self.n_features = len(model.sigma_[0])
        self.n_classes = len(model.classes_)
        self.n_features = len(model.feature_log_prob_[0])

        # jll = safe_sparse_dot(X, (self.feature_log_prob_ - neg_prob).T)
        # jll += self.class_log_prior_ + neg_prob.sum(axis=1)

        # Create class prior probabilities:
        priors = [self.temp('type').format(repr(p)) for p in
                  model.class_log_prior_]
        priors = ', '.join(priors)
        self.priors = self.temp('arr[]').format(type='double', name='priors',
                                                values=priors)

        # Create probabilities:
        # probs = []
        # for prob in model.feature_log_prob_:
        #     tmp = [self.temp('type').format(repr(p)) for p in prob]
        #     tmp = self.temp('arr').format(', '.join(tmp))
        #     probs.append(tmp)
        # probs = ', '.join(probs)
        # self.pos_probs = self.temp('arr[][]').format(type='double',
        #                                              name='posProbs',
        #                                              values=probs)

        # Create negative probabilities:
        neg_prob = np.log(1 - np.exp(model.feature_log_prob_))
        probs = []
        for prob in neg_prob:
            tmp = [self.temp('type').format(repr(p)) for p in prob]
            tmp = self.temp('arr').format(', '.join(tmp))
            probs.append(tmp)
        probs = ', '.join(probs)
        self.neg_probs = self.temp('arr[][]').format(type='double',
                                                     name='negProbs',
                                                     values=probs)

        delta_probs = (model.feature_log_prob_ - neg_prob).T
        probs = []
        for prob in delta_probs:
            tmp = [self.temp('type').format(repr(p)) for p in prob]
            tmp = self.temp('arr').format(', '.join(tmp))
            probs.append(tmp)
        probs = ', '.join(probs)
        self.del_probs = self.temp('arr[][]').format(type='double',
                                                     name='delProbs',
                                                     values=probs)

        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_method(self):
        """
        Build the model method or function.

        Returns
        -------
        :return out : string
            The built method as string.
        """
        indent = 1 if self.language in ['java'] else 0
        return self.temp('method.predict', n_indents=indent,
                         skipping=True).format(**self.__dict__)

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        self.__dict__.update(dict(method=method))
        return self.temp('class').format(**self.__dict__)



# -*- coding: utf-8 -*-

import os
import sys

from sklearn.tree.tree import DecisionTreeClassifier
from sklearn.ensemble.weight_boosting import AdaBoostClassifier
from sklearn.ensemble.forest import RandomForestClassifier
from sklearn.ensemble.forest import ExtraTreesClassifier
from sklearn.svm.classes import LinearSVC
from sklearn.svm.classes import SVC
from sklearn.svm.classes import NuSVC
from sklearn.neighbors.classification import KNeighborsClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.naive_bayes import BernoulliNB


class Porter:

    __version__ = '0.3.2'

    def __init__(self, language="java", method_name='predict', class_name='Tmp',
                 with_details=False):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param language : string (default='java')
            The required syntax ['c', 'go', 'java', 'js'].

        :param method_name : string (default='predict')
            The name of the prediction method.

        :param class_name : string (default='Tmp')
            The name of the environment class.

        :param with_details : bool (default=False)
            Return additional useful information or not.
        """
        self.language = language
        self.method_name = method_name
        self.class_name = class_name
        self.with_details = with_details

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : scikit-learn model object
            An instance of a trained model (e.g. DecisionTreeClassifier).

        Returns
        -------
        :return : string
            The ported model as string.
        """
        cat, name = self.get_model_data(model)
        path = '.'.join([cat, name])  # e.g.: classifier.LinearSVC
        level = -1 if sys.version_info < (3, 3) else 1
        module = __import__(path, globals(), locals(), [name], level)
        klass = getattr(module, name)
        instance = klass(language=self.language, method_name=self.method_name,
                         class_name=self.class_name)
        result = instance.port(model)
        if self.with_details:
            return self.get_details(result)
        return result

    def get_details(self, model):
        """
        Get additional and useful information.

        Parameters
        ----------
        :param model : scikit-learn model object
            An instance of a trained model (e.g. DecisionTreeClassifier()).

        Returns
        -------
        :return data : dict
            language : string
                The target programming language.
            filename : string
                A valid filename.
            compiling_cmd : string
                The command to compile the ported model.
            execution_cmd : string
                The command to execute the ported model (after compiling).
            model : string
                The ported model.
        """

        # Filename:
        filename = self.class_name.lower()
        if self.language == 'java':
            filename = filename.capitalize()
        filename = '%s.%s' % (filename, self.language)

        # Commands:
        comp_cmd = ''  # compiling command
        exec_cmd = ''  # execution command
        if self.language is 'c':
            class_name = self.class_name.lower()
            # gcc tmp.c -o tmp
            comp_cmd = 'gcc %s -o %s' % (filename, class_name)
            # ./tmp
            exec_cmd = os.path.join('.', class_name)
        elif self.language is 'java':
            # javac Tmp.java
            comp_cmd = 'javac %s' % filename
            # java -classpath . Tmp
            class_name = self.class_name.capitalize()
            exec_cmd = 'java -classpath . %s' % class_name
        elif self.language is 'js':
            # node tmp.js
            exec_cmd = 'node %s' % filename
        elif self.language is 'go':
            # TODO: Add go-relevant commands
            pass

        # Result:
        result = {
            'language': self.language,
            'filename': filename,
            'compiling_cmd': comp_cmd,
            'execution_cmd': exec_cmd,
            'model': model,
        }
        return result

    @staticmethod
    def get_model_data(model):
        """
        Get data of the assigned model.

        Parameters
        ----------
        :param model : scikit-learn model object
            An instance of a trained model (e.g. DecisionTreeClassifier()).

        Returns
        -------
        :return cat : string ('regressor', 'classifier')
            The model category.

        :return name : string
            The name of the used algorithm.
        """
        cat = Porter.is_supported_model(model)
        name = type(model).__name__
        return cat, name

    @staticmethod
    def is_supported_classifier(model):
        """
        Check whether the model is a convertible classifier.

        Parameters
        ----------
        :param model : scikit-learn model object
            An instance of a trained model (e.g. DecisionTreeClassifier()).

        Returns
        -------
        :return : bool
            Whether the model is a supported classifier.
        """

        # Default classifiers:
        supported_clfs = (
            DecisionTreeClassifier,
            AdaBoostClassifier,
            RandomForestClassifier,
            ExtraTreesClassifier,
            LinearSVC,
            SVC,
            NuSVC,
            KNeighborsClassifier,
            GaussianNB,
            BernoulliNB,
        )

        # Get version information:
        import sklearn
        version = sklearn.__version__.split('.')
        major, minor = int(version[0]), int(version[1])

        # MLPClassifier was new in version 0.18:
        if major > 0 or (major == 0 and minor >= 18):  # version >= 0.18
            from sklearn.neural_network.multilayer_perceptron import \
                MLPClassifier
            supported_clfs += (MLPClassifier, )

        return isinstance(model, supported_clfs)

    @staticmethod
    def is_supported_regressor(model):
        """
        Check whether the model is a convertible classifier.

        Parameters
        ----------
        :param model : scikit-learn model object
            An instance of a trained model (e.g. DecisionTreeClassifier()).

        Returns
        -------
        :return : bool
            Whether the model is a supported regressor.
        """
        # return isinstance(model, ())
        return False

    @staticmethod
    def is_supported_model(model):
        """
        Check whether the model is a convertible classifier or regressor.

        Parameters
        ----------
        :param model : scikit-learn model object
            An instance of a trained model (e.g. DecisionTreeClassifier()).

        Returns
        -------
        :return : bool
            Whether the model is a supported model.

        See also
        --------
        onl.nok.sklearn.classifier.*, onl.nok.sklearn.regressor.*
        """

        # Is the model a supported classifier?
        if Porter.is_supported_classifier(model):
            return 'classifier'

        # Is the model a supported regressor?
        if Porter.is_supported_regressor(model):
            return 'regressor'

        msg = 'The model is not an instance of '\
              'a supported classifier or regressor.'
        raise ValueError(msg)



# -*- coding: utf-8 -*-

from ..SVC import SVC


class NuSVC(SVC):
    """
    See also
    --------
    sklearn.svm.NuSVC

    http://scikit-learn.org/0.18/modules/generated/sklearn.svm.NuSVC.html
    """

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(NuSVC, self).__init__(language, method_name, class_name)



# -*- coding: utf-8 -*-

from ...Model import Model


class KNeighborsClassifier(Model):
    """
    See also
    --------
    sklearn.neighbors.KNeighborsClassifier

    http://scikit-learn.org/0.18/modules/generated/sklearn.neighbors.KNeighborsClassifier.html
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'java': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'arr[]':    '{type}[] {name} = {{{values}}};',
            'arr[][]':  '{type}[][] {name} = {{{values}}};',
            'indent':   '    ',
        },
        'js': {
            'type':     '{0}',
            'arr':      '[{0}]',
            'arr[]':    'var {name} = [{values}];',
            'arr[][]':  'var {name} = [{values}];',
            'indent':   '    ',
        },
    }
    # @formatter:on

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(KNeighborsClassifier, self).__init__(language, method_name,
                                                   class_name)

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : KNeighborsClassifier
            An instance of a trained KNeighborsClassifier classifier.
        """
        super(self.__class__, self).port(model)

        self.n_classes = len(self.model.classes_)
        self.n_templates = len(self.model._fit_X)
        self.n_features = len(self.model._fit_X[0])
        self.n_neighbors = self.model.n_neighbors

        self.algorithm = self.model.algorithm
        self.power_param = self.model.p

        if self.algorithm is not 'brute':
            from sklearn.neighbors.kd_tree import KDTree
            from sklearn.neighbors.ball_tree import BallTree
            tree = self.model._tree
            if isinstance(tree, (KDTree, BallTree)):
                self.tree = tree

        self.metric = self.model.metric
        if self.model.weights is not 'uniform':
            msg = "Only 'uniform' weights are supported for this classifier."
            raise NotImplementedError(msg)

        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_method(self):
        """
        Build the model method or function.

        Returns
        -------
        :return out : string
            The built method as string.
        """

        # Distance computation
        metric_name = '.'.join(['metric', self.metric])
        distance_comp = self.temp(
            metric_name, n_indents=1, skipping=True)

        # Templates
        temps = []
        for atts in enumerate(self.model._fit_X):
            tmp = [self.temp('type').format(repr(a)) for a in atts[1]]
            tmp = self.temp('arr').format(', '.join(tmp))
            temps.append(tmp)
        temps = ', '.join(temps)
        temps = self.temp('arr[][]').format(
            type='double',
            name='X',
            values=temps,
            n=self.n_templates,
            m=self.n_features)

        # Classes
        classes = self.model._y
        classes = [self.temp('type').format(int(c)) for c in classes]
        classes = ', '.join(classes)
        classes = self.temp('arr[]').format(
            type='int',
            name='y',
            values=classes,
            n=self.n_templates)

        return self.temp('method.predict', n_indents=1, skipping=True).format(
            method_name=self.method_name,
            class_name=self.class_name,
            n_neighbors=self.n_neighbors,
            n_templates=self.n_templates,
            n_features=self.n_features,
            n_classes=self.n_classes,
            distance_computation=distance_comp,
            power=self.power_param,
            X=temps,
            y=classes)

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        return self.temp('class').format(
            class_name=self.class_name,
            method_name=self.method_name,
            method=method,
            n_features=self.n_features)



# -*- coding: utf-8 -*-

from ...Model import Model
import numpy as np

np.set_printoptions(precision=15)


class MLPClassifier(Model):
    """
    See also
    --------
    sklearn.neural_network.MLPClassifier

    http://scikit-learn.org/0.18/modules/generated/sklearn.neural_network.MLPClassifier.html
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'java': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'new_arr':  'new double[{0}]',
            'arr[]':    'double[] {name} = {{{values}}};',
            'arr[][]':  'double[][] {name} = {{{values}}};',
            'arr[][][]': 'double[][][] {name} = {{{values}}};',
            'indent':   '    ',
        },
        'js': {
            'type':     '{0}',
            'arr':      '[{0}]',
            'new_arr':  'new Array({0}).fill(0)',
            'arr[]':    'var {name} = [{values}];',
            'arr[][]':  'var {name} = [{values}];',
            'arr[][][]': 'var {name} = [{values}];',
            'indent':   '    ',
        }
    }
    # @formatter:on

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(MLPClassifier, self).__init__(language, method_name, class_name)

    @property
    def hidden_activation_functions(self):
        """Get list of supported activation functions for the hidden layers."""
        return ['relu', 'identity']  # 'tanh' and 'logistic' fails in tests

    @property
    def output_activation_functions(self):
        """Get list of supported activation functions for the output layer."""
        return ['softmax']  # 'logistic' fails in tests

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : MLPClassifier
            An instance of a trained MLPClassifier classifier.
        """
        super(self.__class__, self).port(model)

        # Activation function ('identity', 'logistic', 'tanh' or 'relu'):
        self.hidden_activation = self.model.activation
        if self.hidden_activation not in self.hidden_activation_functions:
            raise ValueError(('The activation function \'%s\' of the model '
                              'is not supported.') % self.hidden_activation)

        # Output activation function ('softmax' or 'logistic'):
        self.output_activation = self.model.out_activation_
        if self.output_activation not in self.output_activation_functions:
            raise ValueError(('The activation function \'%s\' of the model '
                              'is not supported.') % self.output_activation)

        self.n_layers = self.model.n_layers_
        self.n_hidden_layers = self.model.n_layers_ - 2

        self.n_inputs = len(self.model.coefs_[0])
        self.n_outputs = self.model.n_outputs_

        self.hidden_layer_sizes = self.model.hidden_layer_sizes
        if type(self.hidden_layer_sizes) is int:
            self.hidden_layer_sizes = [self.hidden_layer_sizes]
        self.hidden_layer_sizes = list(self.hidden_layer_sizes)

        self.layer_units = \
            [self.n_inputs] + self.hidden_layer_sizes + [self.model.n_outputs_]

        # Weights:
        self.coefficients = self.model.coefs_

        # Bias:
        self.intercepts = self.model.intercepts_

        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_method(self):
        """
        Build the model method or function.

        Returns
        -------
        :return out : string
            The built method as string.
        """

        # Activations:
        activations = list(self._get_activations())
        activations = ', '.join(['atts'] + activations)
        activations = self.temp('arr[][]').format(
            name='activations', values=activations)

        # Coefficients (weights):
        coefficients = []
        for layer in self.coefficients:
            layer_weights = []
            for weights in layer:
                weights = ', '.join([repr(w) for w in weights])
                layer_weights.append(self.temp('arr').format(weights))
            layer_weights = ', '.join(layer_weights)
            coefficients.append(self.temp('arr').format(layer_weights))
        coefficients = ', '.join(coefficients)
        coefficients = self.temp('arr[][][]').format(
            name='coefficients', values=coefficients)

        # Intercepts (biases):
        intercepts = list(self._get_intercepts())
        intercepts = ', '.join(intercepts)
        intercepts = self.temp('arr[][]').format(
            name='intercepts', values=intercepts)

        return self.temp('method', skipping=True, n_indents=1).format(
            class_name=self.class_name, method_name=self.method_name,
            n_features=self.n_inputs, n_classes=self.n_outputs,
            activations=activations, coefficients=coefficients,
            intercepts=intercepts)

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        hidden_act_type = 'hidden_activation.' + self.hidden_activation
        hidden_act = self.temp(hidden_act_type, skipping=True, n_indents=1)
        output_act_type = 'output_activation.' + self.output_activation
        output_act = self.temp(output_act_type, skipping=True, n_indents=1)
        return self.temp('class').format(
            class_name=self.class_name, method_name=self.method_name,
            method=method, n_features=self.n_inputs,
            hidden_activation_function=hidden_act,
            output_activation_function=output_act)

    def _get_intercepts(self):
        """
        Concatenate all intercepts of the classifier.
        """
        for layer in self.intercepts:
            inter = ', '.join([repr(b) for b in layer])
            yield self.temp('arr').format(inter)

    def _get_activations(self):
        """
        Concatenate the layers sizes of the classifier except the input layer.
        """
        for layer in self.layer_units[1:]:
            yield self.temp('new_arr').format((str(int(layer))))




# -*- coding: utf-8 -*-

import os.path


class Model(object):

    SUPPORTED_METHODS = {}
    TEMPLATES = {}

    def __init__(self, language='java', method_name='predict',
                 class_name='Tmp'):
        self.language = language
        self.method_name = method_name
        self.class_name = class_name
        self.check_support()

    def check_support(self):
        """Check template and programming language support."""

        if self.method_name not in self.__class__.SUPPORTED_METHODS:
            err_msg = ('The given method is not supported '
                       'by the chosen classifier.')
            raise AttributeError(err_msg)

        if self.language not in self.__class__.TEMPLATES.keys():
            err_msg = ('The given programming language is '
                       'not supported for the given method.')
            raise AttributeError(err_msg)

    def indent(self, text, n_indents=1, skipping=False):
        """
        Indent text with single spaces.

        Parameters
        ----------
        :param text : string
            The text which get a specific indentation.
        :param n_indents : int
            The number of indentations.
        :param skipping : boolean
            Whether to skip the initial indentation.

        Returns
        -------
        :return : string
            The indented text.
        """
        lines = text.splitlines()
        space = self.TEMPLATES.get(self.language).get('indent', ' ')

        # Single line:
        if len(lines) == 1:
            if skipping:
                return text.strip()
            return n_indents * space + text.strip()

        # Multiple lines:
        indented_lines = []
        for idx, line in enumerate(lines):
            if skipping and idx is 0:
                indented_lines.append(line)
            else:
                line = n_indents * space + line
                indented_lines.append(line)
        indented_text = '\n'.join(indented_lines)
        return indented_text

    def temp(self, name, templates=None, n_indents=None, skipping=False):
        """
        Get specific template of chosen programming language.

        Parameters
        ----------
        :param name : string
            The key name of the template.
        :param tempaltes : string
            The template with placeholders.
        :param n_indents : int
            The number of indentations.
        :param skipping : boolean
            Whether to skip the initial indentation.

        Returns
        -------
        :return : string
            The wanted template string.
        """
        if templates is None:
            templates = self.TEMPLATES.get(self.language)
        keys = name.split('.')
        key = keys.pop(0).lower()
        template = templates.get(key, None)
        if template is not None:
            if type(template) is str:
                if n_indents is not None:
                    template = self.indent(template, n_indents, skipping)
                return template
            else:
                keys = '.'.join(keys)
                return self.temp(keys, template, skipping=False)
        else:
            # TODO: Replace static path definition: 'classifier'
            class_name = self.__class__.__name__
            path = os.path.join(
                os.path.dirname(__file__), 'classifier', class_name,
                'templates', self.language, name + '.txt')
            if os.path.isfile(path):
                template = open(path, 'r').read()
                if n_indents is not None:
                    template = self.indent(template, n_indents, skipping)
                return template
            else:
                raise AttributeError('Template "%s" not found.' % name)

    def port(self, model):
        """
        Port a trained model in the syntax of a specific programming language.

        Parameters
        ----------
        :param model : scikit-learn model object
            An instance of a trained model (e.g. DecisionTreeClassifier).
        """
        self.model = model



# -*- coding: utf-8 -*-

from ...Model import Model


class GaussianNB(Model):
    """
    See also
    --------
    sklearn.naive_bayes.BernoulliNB

    http://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.BernoulliNB.html#sklearn.naive_bayes.BernoulliNB
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'java': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'arr[]':    '{type}[] {name} = {{{values}}};',
            'arr[][]':  '{type}[][] {name} = {{{values}}};',
            'indent':   '    ',
        },
    }
    # @formatter:on

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(GaussianNB, self).__init__(
            language, method_name, class_name)

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : GaussianNB
            An instance of a trained GaussianNB classifier.
        """
        super(GaussianNB, self).port(model)

        self.n_features = len(model.sigma_[0])
        self.n_classes = len(model.classes_)

        # Create class prior probabilities:
        priors = [self.temp('type').format(repr(c)) for c in
                  self.model.class_prior_]
        priors = ', '.join(priors)
        self.priors = self.temp('arr[]').format(type='double', name='priors',
                                                values=priors)

        # Create sigmas:
        sigmas = []
        for sigma in self.model.sigma_:
            tmp = [self.temp('type').format(repr(s)) for s in sigma]
            tmp = self.temp('arr').format(', '.join(tmp))
            sigmas.append(tmp)
        sigmas = ', '.join(sigmas)
        self.sigmas = self.temp('arr[][]').format(type='double', name='sigmas',
                                                  values=sigmas)

        # Create thetas:
        thetas = []
        for theta in self.model.theta_:
            tmp = [self.temp('type').format(repr(t)) for t in theta]
            tmp = self.temp('arr').format(', '.join(tmp))
            thetas.append(tmp)
        thetas = ', '.join(thetas)
        self.thetas = self.temp('arr[][]').format(type='double', name='thetas',
                                                  values=thetas)

        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_method(self):
        """
        Build the model method or function.

        Returns
        -------
        :return out : string
            The built method as string.
        """
        indent = 1 if self.language in ['java'] else 0
        return self.temp('method.predict', n_indents=indent,
                         skipping=True).format(**self.__dict__)

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        self.__dict__.update(dict(method=method))
        return self.temp('class').format(**self.__dict__)



# -*- coding: utf-8 -*-

from ...Model import Model


class SVC(Model):
    """
    See also
    --------
    sklearn.svm.SVC

    http://scikit-learn.org/0.18/modules/generated/sklearn.svm.SVC.html
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'c': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'arr[]':    '{type} {name}[] = {{{values}}};',
            'arr[][]':  '{type} {name}[{n}][{m}] = {{{values}}};',
            'indent':   '    ',
        },
        'java': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'arr[]':    '{type}[] {name} = {{{values}}};',
            'arr[][]':  '{type}[][] {name} = {{{values}}};',
            'indent':   '    ',
        },
        'js': {
            'type':     '{0}',
            'arr':      '[{0}]',
            'arr[]':    'var {name} = [{values}];',
            'arr[][]':  'var {name} = [{values}];',
            'indent':   '    ',
        },
        'php': {
            'type':     '{0}',
            'arr':      '[{0}]',
            'arr[]':    '${name} = [{values}];',
            'arr[][]':  '${name} = [{values}];',
            'indent':   '    ',
        },
    }
    # @formatter:on

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(SVC, self).__init__(language, method_name, class_name)

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : SVC
            An instance of a trained SVC classifier.
        """
        super(SVC, self).port(model)

        params = model.get_params()
        # Check kernel type:
        kernels = ['linear', 'rbf', 'poly', 'sigmoid']
        if params['kernel'] not in kernels:
            msg = 'The kernel type is not supported.'
            raise ValueError(msg)
        # Check rbf gamma value:
        if params['kernel'] == 'rbf' and params['gamma'] == 'auto':
            msg = ('The classifier gamma value have to '
                   'be set (currently it is \'auto\').')
            raise ValueError(msg)
        self.params = params

        self.svs = model.support_vectors_
        self.n_svs = len(model.support_vectors_[0])
        self.svs_rows = model.n_support_
        self.n_svs_rows = len(model.n_support_)
        self.coeffs = model.dual_coef_
        self.inters = model._intercept_
        self.n_inters = len(model._intercept_)
        self.classes = model.classes_
        self.n_classes = len(model.classes_)

        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_method(self):
        """
        Build the model method or function.

        Returns
        -------
        :return out : string
            The built method as string.
        """
        out = '\n'

        # Number of support vectors:
        n_svs = [self.temp('type').format(repr(v)) for v in self.svs_rows]
        n_svs = ', '.join(n_svs)
        out += self.temp('arr[]').format(type='int', name='n_svs', values=n_svs)
        out += '\n'

        # Support vectors:
        vectors = []
        for vidx, vector in enumerate(self.svs):
            _vectors = [self.temp('type').format(repr(v)) for v in vector]
            _vectors = self.temp('arr').format(', '.join(_vectors))
            vectors.append(_vectors)
        vectors = ', '.join(vectors)
        out += self.temp('arr[][]', skipping=True).format(
            type='double', name='svs', values=vectors,
            n=len(self.svs), m=self.n_svs)
        out += '\n'

        # Coefficients:
        coeffs = []
        for cidx, coeff in enumerate(self.coeffs):
            _coeffs = [self.temp('type').format(repr(c)) for c in coeff]
            _coeffs = self.temp('arr').format(', '.join(_coeffs))
            coeffs.append(_coeffs)
        coeffs = ', '.join(coeffs)
        out += self.temp('arr[][]').format(
            type='double', name='coeffs', values=coeffs,
            n=len(self.coeffs), m=len(self.coeffs[0]))
        out += '\n'

        # Interceptions:
        inters = [self.temp('type').format(repr(i)) for i in self.inters]
        inters = ', '.join(inters)
        out += self.temp('arr[]').format(
            type='double', name='inters', values=inters)
        out += '\n'

        # Classes:
        classes = [self.temp('type').format(repr(c)) for c in self.classes]
        classes = ', '.join(classes)
        out += self.temp('arr[]').format(
            type='int', name='classes', values=classes)
        out += '\n'

        # Kernels:
        if self.params['kernel'] == 'rbf':
            out += self.temp('kernel.rbf').format(
                len(self.svs), self.n_svs, repr(self.params['gamma']))
        elif self.params['kernel'] == 'poly':
            out += self.temp('kernel.poly').format(
                len(self.svs), self.n_svs, repr(self.params['gamma']),
                repr(self.params['coef0']), repr(self.params['degree']))
        elif self.params['kernel'] == 'sigmoid':
            out += self.temp('kernel.sigmoid').format(
                len(self.svs), self.n_svs, repr(self.params['gamma']),
                repr(self.params['coef0']), repr(self.params['degree']))
        elif self.params['kernel'] == 'linear':
            out += self.temp('kernel.linear').format(
                len(self.svs), self.n_svs)
        out += '\n'

        # Decicion:
        out += self.temp('starts').format(self.n_svs_rows)
        out += self.temp('ends').format(self.n_svs_rows)
        out += self.temp('decicions').format(self.n_inters, self.n_svs_rows)
        out += self.temp('classes').format(self.n_inters, self.n_classes)
        wrap = 0 if self.language in ['java', 'js', 'php'] else 1
        out = self.indent(out, n_indents=2-wrap)
        return self.temp('method', n_indents=1-wrap, skipping=True).format(
            method_name=self.method_name, decicion=out)

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        return self.temp('class').format(
            class_name=self.class_name, method_name=self.method_name,
            method=method, n_features=self.n_svs)



# -*- coding: utf-8 -*-

from ...Model import Model


class LinearSVC(Model):
    """
    See also
    --------
    sklearn.svm.LinearSVC

    http://scikit-learn.org/0.18/modules/generated/sklearn.svm.LinearSVC.html
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'c': {
            'type':     ('{0}'),
            'arr':      ('{{{0}}}'),
            'arr[]':    ('double {name}[{n}] = {{{values}}};'),
            'arr[][]':  ('double {name}[{n}][{m}] = {{{values}}};'),
            'indent':   ('    '),
        },
        'go': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'arr[]':    '{name} := []float64{{{values}}}',
            'arr[][]':  '{name} := [][]float64{{{values}}}',
            'indent':   '\t',
        },
        'java': {
            'type':     '{0}',
            'arr':      '{{{0}}}',
            'arr[]':    'double[] {name} = {{{values}}};',
            'arr[][]':  'double[][] {name} = {{{values}}};',
            'indent':   '    ',
        },
        'js': {
            'type':     '{0}',
            'arr':      '[{0}]',
            'arr[]':    'var {name} = [{values}];',
            'arr[][]':  'var {name} = [{values}];',
            'indent':   '    ',
        },
        'php': {
            'type':     '{0}',
            'arr':      '[{0}]',
            'arr[]':    '${name} = [{values}];',
            'arr[][]':  '${name} = [{values}];',
            'indent':   '    ',
        },
        'ruby': {
            'type': '{0}',
            'arr': '[{0}]',
            'arr[]': ('{name} = [{values}]'),
            'arr[][]': ('{name} = [{values}]'),
            'indent': '    ',
        }
    }
    # @formatter:on

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(LinearSVC, self).__init__(language, method_name, class_name)

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : LinearSVC
            An instance of a trained LinearSVC classifier.
        """
        super(self.__class__, self).port(model)

        self.n_features = len(self.model.coef_[0])
        self.n_classes = len(self.model.classes_)

        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_method(self):
        """
        Build the model method or function.

        Returns
        -------
        :return out : string
            The built method as string.
        """

        # Coefficients:
        coefs = []
        for idx, coef in enumerate(self.model.coef_):
            tmp = [self.temp('type').format(repr(c)) for c in coef]
            tmp = self.temp('arr').format(', '.join(tmp))
            coefs.append(tmp)
        coefs = ', '.join(coefs)
        coefs = self.temp('arr[][]').format(
            name='coefs', values=coefs, n=self.n_classes, m=self.n_features)

        # Intercepts:
        inters = self.model.intercept_
        inters = [self.temp('type').format(repr(i)) for i in inters]
        inters = ', '.join(inters)
        inters = self.temp('arr[]').format(
            name='inters', values=inters, n=self.n_classes)

        return self.temp('method', n_indents=1, skipping=True).format(
            name=self.method_name, n_features=self.n_features,
            n_classes=self.n_classes, coefficients=coefs,
            intercepts=inters)

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        return self.temp('class').format(
            class_name=self.class_name, method_name=self.method_name,
            method=method, n_features=self.n_features)



# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

from sklearn_porter import Porter
VERSION = Porter.__version__

path = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(path, 'requirements.txt')) as file:
    all_reqs = file.read().split('\n')
install_requires = [x.strip() for x in all_reqs if 'git+' not in x]

setup(
    name='sklearn-porter',
    packages=find_packages(exclude=["tests.*", "tests"]),
    include_package_data=True,
    version=VERSION,
    description='Transpile trained scikit-learn models to C, Java, JavaScript and others.',
    author='Darius Morawiec',
    author_email='ping@nok.onl',
    url='https://github.com/nok/sklearn-porter/tree/stable',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    install_requires=install_requires,
    keywords=['sklearn', 'scikit-learn'],
    license='MIT',
)



# -*- coding: utf-8 -*-

import os
import argparse

from . import Porter


def main():
    parser = argparse.ArgumentParser(
        description=(
            'Transpile trained scikit-learn models '
            'to C, Java, JavaScript and others. '),
        epilog=(
            'More details on: '
            'https://github.com/nok/sklearn-porter'))
    parser.add_argument(
        '--input', '-i',
        required=True,
        help=(
            'Set the path of an exported model '
            'in pickle (.pkl) format.'))
    parser.add_argument(
        '--output', '-o',
        required=False,
        help='Set the destination directory.')
    parser.add_argument(
        '--language', '-l',
        choices=['c', 'java', 'js', 'go', 'php', 'ruby'],
        default='java',
        required=False,
        help=(
            'Set the target programming language '
            '("c", "java", "js", "go", "php", "ruby").'))
    args = vars(parser.parse_args())

    model_path = str(args['input'])
    if model_path.endswith('.pkl') and os.path.isfile(model_path):
        # Load data:
        from sklearn.externals import joblib
        raw_model = joblib.load(model_path)
        # Port model:
        porter = Porter(language=str(args['language']), with_details=True)
        result = porter.port(raw_model)
        filename = result.get('filename')
        # Define destination path:
        if str(args['output']) != '' and os.path.isdir(str(args['output'])):
            model_path = str(args['output'])
            model_path = os.path.join(model_path, filename)
        else:
            model_path = model_path.split(os.sep)
            # if len(model_path) > 1:
            del model_path[-1]
            model_path += [filename]
            model_path = os.sep.join(model_path)
        # Save ported model:
        with open(model_path, 'w') as f:
            f.write(result.get('model'))
    else:
        raise ValueError('No valid model in pickle format was found.')

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from sklearn.tree.tree import DecisionTreeClassifier

from ...Model import Model


class RandomForestClassifier(Model):
    """
    See also
    --------
    sklearn.ensemble.RandomForestClassifier

    http://scikit-learn.org/0.18/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'c': {
            'if':       '\nif (atts[{0}] {1} {2}) {{',
            'else':     '\n} else {',
            'endif':    '\n}',
            'arr':      '\nclasses[{0}] = {1}\n',
            'indent':   '    ',
            'join':     '; ',
        },
        'java': {
            'if':       '\nif (atts[{0}] {1} {2}) {{',
            'else':     '\n} else {',
            'endif':    '\n}',
            'arr':      '\nclasses[{0}] = {1}\n',
            'indent':   '    ',
            'join':     '; ',
        },
        'js': {
            'if':       '\nif (atts[{0}] {1} {2}) {{',
            'else':     '\n} else {',
            'endif':    '\n}',
            'arr':      '\nclasses[{0}] = {1}\n',
            'indent':   '    ',
            'join':     '; ',
        }
    }
    # @formatter:on

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(RandomForestClassifier, self).__init__(language, method_name,
                                                     class_name)

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : AdaBoostClassifier
            An instance of a trained AdaBoostClassifier model.
        """

        # Check type of base estimators:
        if not isinstance(model.base_estimator, DecisionTreeClassifier):
            msg = "The classifier doesn't support the given base estimator %s."
            raise ValueError(msg, model.base_estimator)

        # Check number of base estimators:
        if not model.n_estimators > 0:
            msg = "The classifier hasn't any base estimators."
            raise ValueError(msg)

        self.model = model
        self.n_classes = model.n_classes_
        self.models = []
        self.n_estimators = 0
        for idx in range(self.model.n_estimators):
            self.models.append(self.model.estimators_[idx])
            self.n_estimators += 1
            self.n_features = self.model.estimators_[idx].n_features_

        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_branches(self, l, r, t, value, features, node, depth):
        """
        Parse and port a single tree model.

        Parameters
        ----------
        :param l : object
            The left children node.
        :param r : object
            The left children node.
        :param t : object
            The decision threshold.
        :param value : object
            The label or class.
        :param features : object
            The feature values.
        :param node : int
            The current node.
        :param depth : int
            The tree depth.

        Returns
        -------
        :return : string
            The ported single tree as function or method.
        """
        out = ''  # returned output
        # ind = '\n' + '    ' * depth
        if t[node] != -2.:
            out += self.temp('if', n_indents=depth).format(
                features[node], '<=', repr(t[node]))
            if l[node] != -1.:
                out += self.create_branches(
                    l, r, t, value, features, l[node], depth + 1)
            out += self.temp('else', n_indents=depth)
            if r[node] != -1.:
                out += self.create_branches(
                    l, r, t, value, features, r[node], depth + 1)
            out += self.temp('endif', n_indents=depth)
        else:
            clazzes = []
            for i, rate in enumerate(value[node][0]):
                clazz = self.temp('arr', n_indents=depth).format(i, int(rate))
                clazzes.append(clazz)
            out += self.temp('join').join(clazzes) + self.temp('join')
        return out

    def create_single_method(self, model_index, model):
        """
        Port a method for a single tree.

        Parameters
        ----------
        :param model_index : int
            The model index.
        :param model : RandomForestClassifier
            The model.

        Returns
        -------
        :return : string
            The created method.
        """
        indices = []
        for i in model.tree_.feature:
            n_features = model.n_features_
            if self.n_features > 1 or (self.n_features == 1 and i >= 0):
                indices.append([str(j) for j in range(n_features)][i])

        tree_branches = self.create_branches(
            model.tree_.children_left, model.tree_.children_right,
            model.tree_.threshold, model.tree_.value, indices, 0, 1)

        suffix = ("{0:0" + str(len(str(self.n_estimators - 1))) + "d}")
        model_index = suffix.format(int(model_index))

        return self.temp('single_method').format(
            model_index, self.method_name, self.n_classes, tree_branches)

    def create_method(self):
        """
        Build the model methods or functions.

        Returns
        -------
        :return out : string
            The built methods as merged string.
        """
        # Generate method or function names:
        fn_names = []
        suffix = ("_{0:0" + str(len(str(self.n_estimators - 1))) + "d}")
        for idx, model in enumerate(self.models):
            fn_name = self.method_name + suffix.format(idx)
            fn_name = self.temp(
                'method_calls', n_indents=2, skipping=True).format(
                idx, self.class_name, fn_name)
            fn_names.append(fn_name)
        fn_names = '\n'.join(fn_names)
        fn_names = self.indent(fn_names, n_indents=1, skipping=True)

        # Generate related trees:
        fns = []
        for idx, model in enumerate(self.models):
            tree = self.create_single_method(idx, model)
            fns.append(tree)
        fns = '\n'.join(fns)

        # Merge generated content:
        indent = 1 if self.language in ['java', 'js'] else 0
        method = self.temp('method').format(
            fns, self.method_name, self.n_estimators, self.n_classes, fn_names)
        method = self.indent(method, n_indents=indent, skipping=True)
        return method

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        return self.temp('class').format(
            self.class_name, self.method_name, method, self.n_features)



# -*- coding: utf-8 -*-

from ...Model import Model


class DecisionTreeClassifier(Model):
    """
    See also
    --------
    sklearn.tree.DecisionTreeClassifier

    http://scikit-learn.org/0.18/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    """

    SUPPORTED_METHODS = ['predict']

    # @formatter:off
    TEMPLATES = {
        'c': {
            'if':       '\nif (atts[{0}] {1} {2}) {{',
            'else':     '\n} else {',
            'endif':    '\n}',
            'arr':      '\nclasses[{0}] = {1}\n',
            'indent':   '    ',
            'join':     '; ',
        },
        'java': {
            'if':       '\nif (atts[{0}] {1} {2}) {{',
            'else':     '\n} else {',
            'endif':    '\n}',
            'arr':      '\nclasses[{0}] = {1}\n',
            'indent':   '    ',
            'join':     '; ',
        },
        'js': {
            'if':       '\nif (atts[{0}] {1} {2}) {{',
            'else':     '\n} else {',
            'endif':    '\n}',
            'arr':      '\nclasses[{0}] = {1}\n',
            'indent':   '    ',
            'join':     '; ',
        },
        'php': {
            'if':       '\nif ($atts[{0}] {1} {2}) {{',
            'else':     '\n} else {',
            'endif':    '\n}',
            'arr':      '\n$classes[{0}] = {1}\n',
            'indent':   '    ',
            'join':     '; ',
        }
    }
    # @formatter:on

    def __init__(
            self, language='java', method_name='predict', class_name='Tmp'):
        super(DecisionTreeClassifier, self).__init__(
            language, method_name, class_name)

    def port(self, model):
        """
        Port a trained model to the syntax of a chosen programming language.

        Parameters
        ----------
        :param model : DecisionTreeClassifier
            An instance of a trained DecisionTreeClassifier classifier.
        """
        super(self.__class__, self).port(model)
        self.n_features = model.n_features_
        self.n_classes = model.n_classes_
        if self.method_name == 'predict':
            return self.predict()

    def predict(self):
        """
        Port the predict method.

        Returns
        -------
        :return: out : string
            The ported predict method.
        """
        return self.create_class(self.create_method())

    def create_branches(self, l, r, t, value, features, node, depth):
        """
        Parse and port a single tree model.

        Parameters
        ----------
        :param l : object
            The left children node.
        :param r : object
            The left children node.
        :param t : object
            The decision threshold.
        :param value : object
            The label or class.
        :param features : object
            The feature values.
        :param node : int
            The current node.
        :param depth : int
            The tree depth.

        Returns
        -------
        :return : string
            The ported single tree as function or method.
        """
        str = ''
        # ind = '\n' + '    ' * depth
        if t[node] != -2.:
            str += self.temp('if', n_indents=depth).format(
                features[node], '<=', repr(t[node]))
            if l[node] != -1.:
                str += self.create_branches(
                    l, r, t, value, features, l[node], depth + 1)
            str += self.temp('else', n_indents=depth)
            if r[node] != -1.:
                str += self.create_branches(
                    l, r, t, value, features, r[node], depth + 1)
            str += self.temp('endif', n_indents=depth)
        else:
            clazzes = []
            for i, rate in enumerate(value[node][0]):
                clazz = self.temp('arr', n_indents=depth).format(i, int(rate))
                clazzes.append(clazz)
            str += self.temp('join').join(clazzes) + self.temp('join')
        return str

    def create_tree(self):
        """
        Parse and build the tree branches.

        Returns
        -------
        :return out : string
            The tree branches as string.
        """
        feature_indices = []
        for i in self.model.tree_.feature:
            n_features = self.n_features
            if self.n_features > 1 or (self.n_features == 1 and i >= 0):
                feature_indices.append([str(j) for j in range(n_features)][i])

        indentation = 1 if self.language in ['java', 'js', 'php'] else 0
        return self.create_branches(
            self.model.tree_.children_left,
            self.model.tree_.children_right,
            self.model.tree_.threshold,
            self.model.tree_.value,
            feature_indices, 0, indentation)

    def create_method(self):
        """
        Build the model method or function.

        Returns
        -------
        :return out : string
            The built method as string.
        """
        indentation = 1 if self.language in ['java', 'js', 'php'] else 0
        branches = self.indent(self.create_tree(), n_indents=1)
        return self.temp('method', n_indents=indentation, skipping=True)\
            .format(method_name=self.method_name, n_features=self.n_features,
                    n_classes=self.n_classes, branches=branches)

    def create_class(self, method):
        """
        Build the model class.

        Returns
        -------
        :return out : string
            The built class as string.
        """
        return self.temp('class').format(
            class_name=self.class_name, method_name=self.method_name,
            n_features=self.n_features, method=method)
